﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Operators;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Algorithms.LocalSearch;
using HeuristicLab.Data;
using HeuristicLab.Optimization;

namespace HeuristicLab.Algorithms.VariableNeighborhoodSearch {
  /// <summary>
  /// A local search improvement operator.
  /// </summary>
  [Item("LocalSearchImprovement", "A local search improvement operator.")]
  [StorableClass]
  public class LocalSearchImprovement: SingleSuccessorOperator, ILocalImprovement {
    [Storable]
    private LocalSearchMainLoop loop;
    
    private /*Constrained*/ValueParameter<IMoveGenerator> MoveGeneratorParameter {
      get { return (/*Constrained*/ValueParameter<IMoveGenerator>)Parameters["MoveGenerator"]; }
    }
    private /*Constrained*/ValueParameter<IMoveMaker> MoveMakerParameter {
      get { return (/*Constrained*/ValueParameter<IMoveMaker>)Parameters["MoveMaker"]; }
    }
    private /*Constrained*/ValueParameter<ISingleObjectiveMoveEvaluator> MoveEvaluatorParameter {
      get { return (/*Constrained*/ValueParameter<ISingleObjectiveMoveEvaluator>)Parameters["MoveEvaluator"]; }
    }
    private ValueParameter<IntValue> MaximumIterationsParameter {
      get { return (ValueParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    private ValueParameter<IntValue> SampleSizeParameter {
      get { return (ValueParameter<IntValue>)Parameters["SampleSize"]; }
    }
    public LookupParameter<IntValue> EvaluatedMovesParameter {
      get { return (LookupParameter<IntValue>)Parameters["EvaluatedMoves"]; }
    }

    public IMoveGenerator MoveGenerator {
      get { return MoveGeneratorParameter.Value; }
      set { MoveGeneratorParameter.Value = value; }
    }
    public IMoveMaker MoveMaker {
      get { return MoveMakerParameter.Value; }
      set { MoveMakerParameter.Value = value; }
    }
    public ISingleObjectiveMoveEvaluator MoveEvaluator {
      get { return MoveEvaluatorParameter.Value; }
      set { MoveEvaluatorParameter.Value = value; }
    }

    [StorableConstructor]
    protected LocalSearchImprovement(bool deserializing) : base(deserializing) { }
    protected LocalSearchImprovement(LocalSearchImprovement original, Cloner cloner)
      : base(original, cloner) {
        this.loop = cloner.Clone(original.loop);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new LocalSearchImprovement(this, cloner);
    }
    public LocalSearchImprovement()
      : base() {
        loop = new LocalSearchMainLoop();

        Parameters.Add(new /*Constrained*/ValueParameter<IMoveGenerator>("MoveGenerator", "The operator used to generate moves to the neighborhood of the current solution."));
        Parameters.Add(new /*Constrained*/ValueParameter<IMoveMaker>("MoveMaker", "The operator used to perform a move."));
        Parameters.Add(new /*Constrained*/ValueParameter<ISingleObjectiveMoveEvaluator>("MoveEvaluator", "The operator used to evaluate a move."));
        Parameters.Add(new ValueParameter<IntValue>("MaximumIterations", "The maximum number of generations which should be processed.", new IntValue(1000)));
        Parameters.Add(new ValueParameter<IntValue>("SampleSize", "Number of moves that MultiMoveGenerators should create. This is ignored for Exhaustive- and SingleMoveGenerators.", new IntValue(100)));
        Parameters.Add(new LookupParameter<IntValue>("EvaluatedMoves", "The number of evaluated moves."));
    }

    public void Parameterize(IProblem problem) {
      UpdateMoveOperators(problem);
    }

    private void UpdateMoveOperators(IProblem problem) {
     /* IMoveGenerator oldMoveGenerator = MoveGenerator;
      MoveGeneratorParameter.ValidValues.Clear();
      if (problem != null) {
        foreach (IMoveGenerator generator in problem.Operators.OfType<IMoveGenerator>().OrderBy(x => x.Name))
          MoveGeneratorParameter.ValidValues.Add(generator);

        foreach (IMoveMaker maker in problem.Operators.OfType<IMoveMaker>().OrderBy(x => x.Name))
          MoveMakerParameter.ValidValues.Add(maker);

        foreach (ISingleObjectiveMoveEvaluator evaluator in problem.Operators.OfType<ISingleObjectiveMoveEvaluator>().OrderBy(x => x.Name))
          MoveEvaluatorParameter.ValidValues.Add(evaluator);
      }

      if (oldMoveGenerator != null) {
        IMoveGenerator newMoveGenerator = MoveGeneratorParameter.ValidValues.FirstOrDefault(x => x.GetType() == oldMoveGenerator.GetType());
        if (newMoveGenerator != null) MoveGenerator = newMoveGenerator;
      }
      if (MoveGenerator == null) {
        ClearMoveParameters();
      }

      IMoveMaker oldMoveMaker = MoveMaker;
      if (oldMoveMaker != null) {
        IMoveMaker mm = MoveMakerParameter.ValidValues.FirstOrDefault(x => x.GetType() == oldMoveMaker.GetType());
        if (mm != null) MoveMaker = mm;
      }

      ISingleObjectiveMoveEvaluator oldMoveEvaluator = MoveEvaluator;
      if (oldMoveEvaluator != null) {
        ISingleObjectiveMoveEvaluator me = MoveEvaluatorParameter.ValidValues.FirstOrDefault(x => x.GetType() == oldMoveEvaluator.GetType());
        if (me != null) MoveEvaluator = me;
      }*/
    }

    private void ClearMoveParameters() {
      /*MoveMakerParameter.ValidValues.Clear();
      MoveEvaluatorParameter.ValidValues.Clear();*/
    }

    public override IOperation Apply() {
      Scope subScope = new Scope();
      Scope individual = new Scope();

      foreach (Variable var in ExecutionContext.Scope.Variables) {
        individual.Variables.Add(var);
      }
      subScope.SubScopes.Add(individual);

      ExecutionContext.Scope.SubScopes.Add(subScope);
      int index = subScope.Parent.SubScopes.IndexOf(subScope);

      SubScopesProcessor processor = new SubScopesProcessor();
      SubScopesRemover remover = new SubScopesRemover();

      remover.RemoveAllSubScopes = false;
      remover.SubScopeIndexParameter.Value = new IntValue(index);

      for (int i = 0; i < index; i++) {
        processor.Operators.Add(new EmptyOperator());
      }

      loop.MoveGeneratorParameter.Value = MoveGeneratorParameter.Value;
      if (loop.MoveGeneratorParameter.Value != null && loop.MoveGeneratorParameter.Value.Parameters.ContainsKey("SampleSize")) {
        IParameter parameter = loop.MoveGeneratorParameter.Value.Parameters["SampleSize"];
        if(parameter is IValueParameter)
          (parameter as IValueParameter).Value = SampleSizeParameter.Value;
      }
      loop.MoveEvaluatorParameter.Value = MoveEvaluatorParameter.Value;
      loop.MoveMakerParameter.Value = MoveMakerParameter.Value;
      loop.MaximumIterationsParameter.Value = MaximumIterationsParameter.Value;

      processor.Operators.Add(loop);
      processor.Successor = remover;

      IOperation next = base.Apply();
      if (next as ExecutionContext != null) {
        remover.Successor = (next as ExecutionContext).Operator;
      }

      return ExecutionContext.CreateOperation(processor);
    }
  }
}
